/******************************************************/
/******************************************************/
/*****************    Main gc thread  *****************/
/******************************************************/
/******************************************************/


/*
get_new_meminfos(mv,num_new_allocations):
  for i = 0 to num_new_allocations 
      from = arb_read(
                 mv.arb_reader,
                 sizeof(struct meminfo_minimal)
              )
      to   = get_pool(mv.meminfo_pool)

      to.start    = from.start
      to.end      = from.start+from.size
      to.marked   = false
      to.iteriors = {}

      put(to, mv.all_meminfos)
      put_hash(mv.hash_table,to.start,to)
*/

void get_new_meminfos(mark_variables_t *mv,int num_new_allocations){
  int i;
  for(i=0;i<num_new_allocations;i++){
    meminfo_minimal_t *from = stack_read(mv->reader,sizeof(meminfo_minimal_t));
    //meminfo_t         *to   = poolmem_alloc(mv->poolmem,sizeof(meminfo_t));
    meminfo_t         *to   = pool_get(mv->meminfo_pool,false);
    size_t             size;

    //fprintf(stderr,"i/num: %d/%d\n",i,num_new_allocations);
    assert(from != NULL);
    assert(to != NULL);

    to->iteriors      = NULL;
    to->iterior_owner = NULL;

    minimal_read(from,mv->heap_start,mv->atomic_heap_start,(void**)&to->start,&size,&to->is_atomic);

    assert(to->start != NULL);
    assert(size > 0);

    //fprintf(stderr,"got new. Size: %d, type: %d\n",size,to->is_atomic);

    to->end      = to->start+size;
    to->marked   = false;

    to->next = mv->all_meminfos;
    mv->all_meminfos = to;

    hash_put(mv->hash,&to->hash_elem,to->start);
  }
}


/*
take_heap_snapshot(heap):
  \new{if at_least_one_second_since_last_full_copy(current_heap):}
     \new{size = heap_size}
  \new{else:}
     \new{size = heap.mheap.freemem - heap.start}

  for i = 0 \new{size}:
      snapshot_mem[i] = heap.start[i]
*/

static void take_heap_snapshot(tar_heap_t *heap){
  size_t size =
    heap->take_full_snapshot 
    ? heap_size 
    : GET_CONT_HEAP_SIZE(heap->mheap,heap_size);
  memcpy(snapshot_mem,GET_RAW_MEM(heap->mheap),size);
}




/*
prepare_new_gc(mv, heap):
  mv.all_meminfos         = heap.all_meminfos
  mv.heap_start           = heap.start
  mv.heap_end             = heap.end
  mv.snapshot_heap_offset = snapshot_heap - heap.start

  arbo_switch(current_heap.ringbuffer)
  heap.arb_writer         = heap.ringbuffer.writer
  mv.arb_reader           = heap.ringbuffer.reader
*/


static void prepare_new_gc(mark_variables_t *mv,tar_heap_t *heap){
  mv->all_meminfos        = heap->all_meminfos;
  mv->heap_start          = GET_RAW_MEM(heap->mheap);
  mv->heap_end            = GET_CONT_HEAP_END(heap->mheap,heap_size);
  //mv->heap_end            = mv->heap_start+heap_size;
  mv->atomic_heap_start   = GET_RAW_MEM(atomic_heap);
  mv->atomic_heap_end     = GET_CONT_HEAP_END(atomic_heap,atomic_heap_size);
  //mv->atomic_heap_end     = mv->atomic_heap_start+atomic_heap_size;
  mv->snapshot_mem_offset = snapshot_mem - mv->heap_start;
  mv->hash                = heap->hash;

  twostacks_switch(heap->twostacks);
  heap->writer           = heap->twostacks->writer;
  mv->reader             = heap->twostacks->reader;  

  mv->num_new_allocations = stack_size(mv->reader) / sizeof(meminfo_minimal_t);
}


/*
gc_thread():
  mv              = sys_alloc(sizeof(struct mark_variables))}
  mv.meminfo_pool = init_pool(sizeof(struct meminfo))}

  while true:
     wait(collector_ready)
       heap = current_heap 
       reset_roots()
       take_root_snapshot(global_vars_start, global_vars_end)
       take_heap_snapshot(heap)
       prepare_new_gc(mv, heap)
       num_new_allocations = arb_space(mv.arb_reader) / sizeof(struct meminfo_minimal)
     signal(heap.dsp_function_can_run)

     if num_new_allocations > 0:
        set_lower_priority()
          get_new_meminfos(mv,num_new_allocations)
          run_mark(mv, root_start, root_end)
          run_sweep()
        set_higher_priority()
*/

static meminfo_t *unfreed_meminfos=NULL;
static meminfo_t *unfreed_atomic_meminfos=NULL;

static void mark_and_sweep_thread(void){
  set_mark_thread_prio();
  
  while(true){

    ///////////////
    mark_variables_t *mv=semaphore_wait(mark_and_sweep_ready);


    ///////////////
    get_new_meminfos(mv,mv->num_new_allocations);


    ///////////////
    run_mark(mv, roots_start, roots_end);


    ///////////////
    reset_roots();


    ///////////////
    if(unfreed_atomic_meminfos!=NULL){
      meminfo_t *last = unfreed_atomic_meminfos;
      meminfo_t *first = last->next;
      last->next = mv->heap->all_meminfos;
      mv->heap->all_meminfos = first;
      unfreed_atomic_meminfos = NULL;
    }


    ///////////////
    run_sweep(mv);


    ///////////////
    if(unfreed_meminfos!=NULL){
      meminfo_t *last  = unfreed_meminfos;
      pool_put_list(mv->meminfo_pool,last);
      unfreed_meminfos = NULL;
    }

  }
}


static mark_variables_t *global_mv;

static void free_meminfo_list(meminfo_t *first,meminfo_t *last,bool collector_is_running){
  last->next = first;

  if(collector_is_running){

    while(unfreed_meminfos != NULL){
      usleep(1000000/100);
    }

    unfreed_meminfos=last;

  }else{
    pool_put_list(global_mv->meminfo_pool,last);
  }
}

static void free_atomic_meminfo_list(meminfo_t *first,meminfo_t *last,bool collector_is_running){

  if(collector_is_running==false || all_heaps==NULL){

    meminfo_t *mem=first;
    while(mem!=NULL){
      FREE_MEM(atomic_heap, mem->start, mem->end - mem->start);
      mem=mem->next;
    }
    pool_put_list(global_mv->meminfo_pool,last);

  }else{

    last->next = first;
    while(unfreed_meminfos != NULL){
      usleep(1000000/100);
    }
    unfreed_atomic_meminfos=last;

  }
}


static void gc_thread(void){
  mark_variables_t mv={0};
  global_mv=&mv;

  mv.meminfo_pool = pool_create(sizeof(meminfo_t),POOL_SIZE_FACTOR*heap_size); //512*1024);
  //mv.poolmem      = poolmem_create(POOL_SIZE_FACTOR*heap_size,NULL);

  set_snapshot_thread_prio();

  while(true){
    tar_heap_t *heap=semaphore_wait(collector_ready);
    assert(semaphore_is_waiting(mark_and_sweep_ready)==true);
    mv.heap = heap;

    take_roots_snapshot(heap);
    take_heap_snapshot(heap);
    prepare_new_gc(&mv,heap);

    semaphore_signal(heap->audio_function_can_run);

    if(mv.num_new_allocations>0){
      semaphore_signal_val(mark_and_sweep_ready,&mv);
    }else{
      reset_roots();
    }

    //    while(semaphore_is_waiting(mark_and_sweep_ready)==false)
    //   usleep(5000);
  }
}



