
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <inttypes.h>
#include <unistd.h>


typedef unsigned long long cycles_t;

#if defined(__x86_64__)
typedef long unsigned int jack_time_t;
#else
typedef long long unsigned int jack_time_t;
#endif



static jack_time_t __jack_cpu_mhz = 0;






static inline cycles_t get_cycles (void)
{
        unsigned long long ret;
        __asm__ __volatile__("rdtsc" : "=A" (ret));
        return ret;
}


jack_time_t jack_get_microseconds_from_cycles (void) {
        return get_cycles() / __jack_cpu_mhz;
}

double cycle_to_time(cycles_t cycles){
  return cycles/((double)__jack_cpu_mhz*1000000.0);
}

float cycle_to_microtime(cycles_t cycles){
  return (float)(cycles/((double)__jack_cpu_mhz*1000.0));
}


void jack_error(char *ai){
  fprintf(stderr,ai);
}

/*
 * This is another kludge.  It looks CPU-dependent, but actually it
 * reflects the lack of standards for the Linux kernel formatting of
 * /proc/cpuinfo.
 */

jack_time_t jack_get_mhz (void)
{
  FILE *f = fopen("/proc/cpuinfo", "r");
  if (f == 0)
    {
      perror("can't open /proc/cpuinfo\n");
      exit(1);
    }
  
  for ( ; ; )
    {
      jack_time_t mhz;
      int ret;
      char buf[1000];
      
      if (fgets(buf, sizeof(buf), f) == NULL) {
	jack_error ("FATAL: cannot locate cpu MHz in "
		    "/proc/cpuinfo\n");
	exit(1);
      }
      
#if defined(__powerpc__)
      ret = sscanf(buf, "clock\t: %" SCNu64 "MHz", &mhz);
#elif defined( __i386__ ) || defined (__hppa__)  || defined (__ia64__) || \
  defined(__x86_64__)
      ret = sscanf(buf, "cpu MHz         : %" SCNu64, &mhz);
#elif defined( __sparc__ )
      ret = sscanf(buf, "Cpu0Bogo        : %" SCNu64, &mhz);
#elif defined( __mc68000__ )
      ret = sscanf(buf, "Clocking:       %" SCNu64, &mhz);
#elif defined( __s390__  )
      ret = sscanf(buf, "bogomips per cpu: %" SCNu64, &mhz);
#else /* MIPS, ARM, alpha */
      ret = sscanf(buf, "BogoMIPS        : %" SCNu64, &mhz);
#endif
      
      if (ret == 1)
	{
	  fclose(f);
	  return (jack_time_t)mhz;
	}
    }
}


void jack_init_time ()
{
        __jack_cpu_mhz = jack_get_mhz ();
}


#ifdef TIMINGTEST
int main(){
  jack_init_time();
  printf("mzh: %llu\n",__jack_cpu_mhz);
  cycles_t start=get_cycles();
  printf("start-time: %llu\n",start);
  sleep(1);
  cycles_t end=get_cycles();
  printf("end-time: %llu, diff: %llu\n",end,end-start);
  printf("len: %f\n",(float)cycle_to_time(end-start));
  return 0;
}
#endif
